import signal, os
import asyncio
import traceback

class Worker(object):
    """
    Wrapper for async function for single ARENA tasks.
    """
    def __init__(self, event_loop, func, event=None, *args, **kwargs):
        self.event_loop = event_loop
        self.func = func
        self.event = event
        self.args = args
        self.kwargs = kwargs

    async def run(self):
        if self.event: await self.event.wait()
        try:
            self.func(*self.args, **self.kwargs)
        except Exception as e:
            func_name = self.func.__name__
            Worker.print_traceback(func_name)

            # disconnect and stop running the program
            self.event_loop.shutdown_wrapper()

    async def sleep(self, interval):
        await asyncio.sleep(interval)

    @classmethod
    def print_traceback(cls, func_name):
        print()
        print(f"Exception thrown in {func_name}()! Terminating {func_name}...")
        # ignore traceback from event_loop.py
        traceback_data = traceback.format_exc().splitlines()
        traceback_str = "\n".join(traceback_data[:1]+traceback_data[3:])
        print(traceback_str)
        print()



from .attribute import Attribute

class JitsiVideo(Attribute):
    """
    Jitsi Video Attribute. For video source.
    Usage: jitsi_video=JitsiVideo(...)
    """
    def __init__(self, jitsiId=None, displayName=None):
        # Since jitsiId overrides displayName, do not set None arguments
        kwargs = {}
        if jitsiId is not None:
            kwargs['jitsiId'] = jitsiId
        if displayName is not None:
            kwargs['displayName'] = displayName
        super().__init__(**kwargs)



#!/usr/bin/env python3

from arena import Scene

def main():
    scene = Scene()
    print(f"ARENA Username: {scene.username}")
    print(f"ARENA Token: {scene.remote_auth_token['token']}")

if __name__ == "__main__":
    main()



from .worker import Worker

class LazyWorker(Worker):
    """
    Wrapper for an ARENA task that runs after an interval (ms).
    """
    def __init__(self, event_loop, func, event=None, interval=1000, *args, **kwargs):
        super().__init__(event_loop, func, event, *args, **kwargs)
        self.interval = interval / 1000 # ms -> s

    async def run(self):
        if self.event: await self.event.wait()
        await self.sleep(self.interval)
        try:
            self.func(*self.args, **self.kwargs)
        except Exception as e:
            func_name = self.func.__name__
            Worker.print_traceback(func_name)
            return



from .attribute import Attribute

class TextInput(Attribute):
    """
    TextInput Attribute.
    Usage: textinput=TextInput(...)
    """
    def __init__(self,
                title="Text Input",
                label="Input text below (max is 140 characters)",
                placeholder="Type here",
                on="mousedown"):
        super().__init__(title=title[:140], label=label[:140], placeholder=placeholder[:140], on=on)



import socket
import asyncio
import paho.mqtt.client as mqtt
from .async_worker import AsyncWorker

class AsyncioMQTTHelper:
    """Adapted from: https://github.com/eclipse/paho.mqtt.python/blob/master/examples/loop_asyncio.py"""
    def __init__(self, event_loop, client):
        self.event_loop = event_loop
        self.loop = self.event_loop.loop

        self.client = client

        self.client.on_socket_open = self.on_socket_open
        self.client.on_socket_close = self.on_socket_close
        self.client.on_socket_register_write = self.on_socket_register_write
        self.client.on_socket_unregister_write = self.on_socket_unregister_write

    def on_socket_open(self, client, userdata, sock):
        def cb():
            client.loop_read()

        self.loop.add_reader(sock, cb)
        self.misc_loop_worker = AsyncWorker(
                                self.event_loop,
                                func=self.misc_loop,
                                event=None,
                            )
        self.misc = self.event_loop.add_task(self.misc_loop_worker)

    def on_socket_close(self, client, userdata, sock):
        self.loop.remove_reader(sock)
        # self.misc.cancel()

    def on_socket_register_write(self, client, userdata, sock):
        def cb():
            client.loop_write()

        self.loop.add_writer(sock, cb)

    def on_socket_unregister_write(self, client, userdata, sock):
        self.loop.remove_writer(sock)

    async def misc_loop(self):
        while self.client.loop_misc() == mqtt.MQTT_ERR_SUCCESS:
            try:
                await asyncio.sleep(1)
            except asyncio.CancelledError:
                break



from .arena_object import Object

class Text(Object):
    """
    Class for Text in the ARENA.
    """
    object_type = "text"

    def __init__(self, text="placeholder text", **kwargs):
        super().__init__(object_type=Text.object_type, text=text, **kwargs)



from ..utils import Utils
from .attribute import Attribute
from .position import Position

class Landmark(Attribute):
    """
    Landmark Attribute.
    Usage: Animation(...)
    """
    def __init__(self, **kwargs):
        if "offsetPosition" in kwargs:
            if isinstance(kwargs["offsetPosition"], tuple) or isinstance(kwargs["offsetPosition"], list):
                kwargs["offsetPosition"] = Utils.tuple_to_string(kwargs["offsetPosition"])
            # can get away with using Position.to_str() even if animation is rotation (euler) or scale
            # since we just want the 3 values to be a space separated string
            elif isinstance(kwargs["offsetPosition"], Position):
                kwargs["offsetPosition"] = kwargs["offsetPosition"].to_str()
        super().__init__(**kwargs)



from .arena_object import Object

class Box(Object):
    """
    Class for Box in the ARENA.
    """
    object_type = "box"

    def __init__(self, **kwargs):
        super().__init__(object_type=Box.object_type, **kwargs)

class Cube(Object):
    """
    Class for Cube in the ARENA.
    """
    object_type = "cube"

    def __init__(self, **kwargs):
        print("[DEPRECATED]", "Cube(...) will be deprecated soon, please use Box(...) instead!")
        super().__init__(object_type=Cube.object_type, **kwargs)



import argparse
import asyncio
import json
import os
import random
import socket
import ssl
import sys

import paho.mqtt.client as mqtt

from .auth import ArenaAuth
from .event_loop import *


class ArenaMQTT(object):
    """
    Wrapper around Paho MQTT client and EventLoop.
    """

    scene = None
    device = None
    auth = None

    def __init__(
                self,
                realm = "realm",
                network_latency_interval = 10000,  # run network latency update every 10s
                on_msg_callback = None,
                end_program_callback = None,
                video = False,
                debug = False,
                **kwargs
            ):
        if os.environ.get("MQTTH"):
            self.host = os.environ["MQTTH"]
        elif "host" in kwargs and kwargs["host"]:
            self.host = kwargs["host"]
            print("Cannot find MQTTH environmental variable, using input parameter instead.")
        else:
            sys.exit("mqtt host argument (host) is unspecified or None, aborting...")

        if os.environ.get("REALM"):
            self.realm = os.environ["REALM"]
        elif "realm" in kwargs and kwargs["realm"]:
            self.realm = kwargs["realm"]
            print("Cannot find REALM environmental variable, using input parameter instead.")
        else:
            # Use default "realm" until multiple realms exist, avoids user confusion.
            self.realm = realm

        self.debug = debug

        print("=====")
        # do user auth
        self.username = None
        token = None
        self.remote_auth_token = {}  # provide reference for downloaded token
        self.auth = ArenaAuth()
        if os.environ.get("ARENA_USERNAME") and os.environ.get("ARENA_PASSWORD"):
            # auth 1st: use passed in env var
            self.username = os.environ["ARENA_USERNAME"]
            token = os.environ["ARENA_PASSWORD"]
            self.auth.store_environment_auth(self.username, token)
        else:
            if self.scene:
                local = self.auth.check_local_auth()
            elif self.device:
                local = self.auth.authenticate_device(self.host)
            if local and "username" in local and "token" in local:
                # auth 2nd: use locally saved token
                self.username = local["username"]
                token = local["token"]
            else:
                if self.scene:
                    # auth 3rd: use the user account online
                    self.username = self.auth.authenticate_user(self.host)

        if os.environ.get("NAMESPACE"):
            self.namespace = os.environ["NAMESPACE"]
        elif "namespace" not in kwargs or ("namespace" in kwargs and kwargs["namespace"] is None):
            self.namespace = self.username
        else:
            self.namespace = kwargs["namespace"]

        # account for mqtt.arenaxr.org
        if self.host == "arenaxr.org" and not "mqtt." in self.host:
            self.host = "mqtt." + self.host
        self.mqttc_id = "pyClient-" + self.generate_client_id()

        # fetch host config
        print("Fetching ARENA configuration...")
        self.config_url = f"https://{self.host}/conf/defaults.json"
        self.config_data = json.loads(self.auth.urlopen(self.config_url))

        # set up topic variables
        if self.scene:
            self.namespaced_target =  f"{self.namespace}/{self.scene}"
            self.root_topic = f"{self.realm}/s/{self.namespaced_target}"
        elif self.device:
            self.namespaced_target =  f"{self.namespace}/{self.device}"
            self.root_topic = f"{self.realm}/d/{self.namespaced_target}"
        self.subscribe_topic = f"{self.root_topic}/#"   # main topic for entire target
        self.latency_topic = self.config_data["ARENADefaults"]["latencyTopic"] # network graph latency update
        self.ignore_topic = f"{self.root_topic}/{self.mqttc_id}/#" # ignore own messages

        self.mqttc = mqtt.Client(
            self.mqttc_id, clean_session=True
        )

        if self.scene and (not self.username or not token):
            # do scene auth by user
            data = self.auth.authenticate_scene(
                self.host, self.realm, self.namespaced_target, self.username, video)
            if "username" in data and "token" in data:
                self.username = data["username"]
                token = data["token"]
                self.remote_auth_token = data
        self.mqttc.username_pw_set(username=self.username, password=token)
        print("=====")

        # set up callbacks
        self.on_msg_callback = on_msg_callback
        self.end_program_callback = end_program_callback

        self.event_loop = EventLoop(self.disconnect)

        aioh = AsyncioMQTTHelper(self.event_loop, self.mqttc)

        # have all tasks wait until mqtt client is connected before starting
        self.mqtt_connect_evt = asyncio.Event()
        self.mqtt_connect_evt.clear()

        # set paho mqtt callbacks
        self.mqttc.on_connect = self.on_connect
        self.mqttc.on_disconnect = self.on_disconnect

        # add main message processing + callbacks loop to tasks
        self.run_async(self.process_message)

        # update network latency every network_latency_interval secs
        self.run_forever(self.network_latency_update,
                         interval_ms=network_latency_interval)

        self.msg_queue = asyncio.Queue()

        # connect to mqtt broker
        if "port" in kwargs:
            port = kwargs["port"]
        else:
            port = 8883 # ARENA broker TLS 1.2 connection port
        if self.auth.verify(self.host):
            self.mqttc.tls_set()
        else:
            self.mqttc.tls_set_context(ssl._create_unverified_context())
            self.mqttc.tls_insecure_set(True)
        try:
            self.mqttc.connect(self.host, port=port)
        except Exception as err:
            print(f'MQTT connect error to {self.host}, port={port}: {err}')
        self.mqttc.socket().setsockopt(socket.SOL_SOCKET, socket.SO_SNDBUF, 2048)


    def parse_cli(self):
        """
        Reusable command-line options to give apps flexible options to avoid hard-coding locations.
        """
        parser = argparse.ArgumentParser(description=("ARENA-py Application CLI"))
        parser.add_argument("-mh", "--mqtth", type=str,
                            help="MQTT host to connect to")
        parser.add_argument("-n", "--namespace", type=str,
                            help="Namespace of scene")
        parser.add_argument("-s", "--scene", type=str,
                            help="Scene to publish and listen to")
        parser.add_argument("-d", "--device", type=str,
                            help="Device to publish and listen to")
        parser.add_argument("-p", "--position", nargs=3, type=float, default=(0, 0, 0),
                            help="App position as cartesian.x cartesian.y cartesian.z")
        parser.add_argument("-r", "--rotation", nargs=3, type=float, default=(0, 0, 0),
                            help="App rotation as euler.x euler.y euler.z")
        parser.add_argument("-c", "--scale", nargs=3, type=float, default=(1, 1, 1),
                            help="App scale in meters")
        args = parser.parse_args()
        app_position = tuple(args.position)
        app_rotation = tuple(args.rotation)
        app_scale = tuple(args.scale)
        return {
            "mqtth": args.mqtth,
            "namespace": args.namespace,
            "scene": args.scene,
            "device": args.device,
            "position": app_position,
            "rotation": app_rotation,
            "scale": app_scale,
        }


    def generate_client_id(self):
        """Returns a random 6 digit id"""
        return str(random.randrange(100000, 999999))

    def network_latency_update(self):
        """Update client latency in $NETWORK/latency"""
        # publish empty message with QoS of 2 to update latency
        self.mqttc.publish(self.latency_topic, "", qos=2)

    def run_once(self, func=None, **kwargs):
        """Runs a user-defined function on startup"""
        if func is not None:
            w = SingleWorker(self.event_loop, func, self.mqtt_connect_evt, **kwargs)
            self.event_loop.add_task(w)
        else:
            # if there is no func, we are in a decorator
            def _run_once(func):
                self.run_once(func, **kwargs)
                return func
            return _run_once

    def run_after_interval(self, func=None, interval_ms=1000, **kwargs):
        """Runs a user-defined function after a interval_ms milliseconds"""
        if func is not None:
            if interval_ms < 0:
                print("Invalid interval! Defaulting to 1000ms")
                interval_ms = 1000
            w = LazyWorker(self.event_loop, func, self.mqtt_connect_evt, interval_ms, **kwargs)
            self.event_loop.add_task(w)
        else:
            # if there is no func, we are in a decorator
            def _run_after_interval(func):
                self.run_after_interval(func, interval_ms, **kwargs)
                return func
            return _run_after_interval

    def run_async(self, func=None, **kwargs):
        """Runs a user defined aynscio function"""
        if func is not None:
            w = AsyncWorker(self.event_loop, func, self.mqtt_connect_evt, **kwargs)
            self.event_loop.add_task(w)
        else:
            # if there is no func, we are in a decorator
            def _run_async(func):
                self.run_async(func, **kwargs)
                return func
            return _run_async

    def run_forever(self, func=None, interval_ms=1000, **kwargs):
        """Runs a function every interval_ms milliseconds"""
        if func is not None:
            if interval_ms < 0:
                print("Invalid interval! Defaulting to 1000ms")
                interval_ms = 1000
            t = PersistentWorker(self.event_loop, func, self.mqtt_connect_evt, interval_ms, **kwargs)
            self.event_loop.add_task(t)
        else:
            # if there is no func, we are in a decorator
            def _run_forever(func):
                self.run_forever(func, interval_ms, **kwargs)
                return func
            return _run_forever

    def run_tasks(self):
        """Run event loop"""
        print("Connecting to the ARENA... ", end="")
        self.event_loop.run()

    def stop_tasks(self):
        """Stop event loop"""
        self.event_loop.stop()

    async def sleep(self, interval_ms):
        """Public function for sleeping in async functions"""
        await asyncio.sleep(interval_ms / 1000)

    def on_connect(self, client, userdata, flags, rc):
        """Paho MQTT client on_connect callback"""
        if rc == 0:
            self.mqtt_connect_evt.set()

            # listen to all messages in scene
            client.subscribe(self.subscribe_topic)
            client.message_callback_add(self.subscribe_topic, self.on_message)

            print("Connected!")
            print("=====")
        else:
            print(f"Connection error! Result code: {rc}")

    def on_message(self, client, userdata, msg):
        # ignore own messages
        if mqtt.topic_matches_sub(self.ignore_topic, msg.topic):
            return

        self.msg_queue.put_nowait(msg)

    async def process_message(self):
        raise NotImplementedError("Must override process_message")

    def on_disconnect(self, client, userdata, rc):
        """Paho MQTT client on_disconnect callback"""
        if rc == 0:
            print("Disconnected from the ARENA!")
        else:
            print(f"Disconnect error! Result code: {rc}")

    def disconnect(self):
        """Disconnects Paho MQTT client"""
        if self.end_program_callback:
            self.end_program_callback(self)
        self.mqttc.disconnect()

    def message_callback_add(self, sub, callback):
        """Subscribes to new topic and adds callback"""
        self.mqttc.subscribe(sub)
        self.mqttc.message_callback_add(sub, callback)

    def message_callback_remove(self, sub):
        """Unsubscribes to topic and removes callback"""
        self.mqttc.unsubscribe(sub)
        self.mqttc.message_callback_remove(sub)



#!/usr/bin/env python3
"""
arena-py-sub: Helper script for ARENA MQTT CLI subscribe
"""
import json
import os
import sys

def main():
    sargs = sys.argv
    args = []
    for arg in sargs:
        try:
            json.loads(arg)
            # preserve shell-stripped quotes
            if "\"" in arg:
                args.append(f"'{arg}'")
            else:
                args.append(f"\"{arg}\"")
        except:
            args.append(arg)
    args[0] = "python3 -m arena -a sub"
    os.system(" ".join(args))

if __name__ == "__main__":
    main()



import json
import os
import re
import sys
from datetime import datetime
from inspect import signature

from .arena_mqtt import ArenaMQTT
from .attributes import *
from .events import *
from .objects import *
from .utils import *


class Scene(ArenaMQTT):
    """
    Gives access to an ARENA scene.
    Can create and execute various user-defined functions/tasks.

    :param str host: Hostname of the MQTT broker (required).
    :param str realm: Reserved topic fork for future use (optional).
    :param str namespace: Username of authenticated user or other namespace (automatic).
    :param str scene: The name of the scene, without namespace (required).
    """

    def __init__(
                self,
                realm = "realm",
                network_latency_interval = 10000,  # run network latency update every 10s
                on_msg_callback = None,
                new_obj_callback = None,
                user_join_callback = None,
                user_left_callback = None,
                delete_obj_callback = None,
                end_program_callback = None,
                video = False,
                debug = False,
                cli_args = False,
                **kwargs
            ):
        if cli_args:
            self.args = self.parse_cli()
            if self.args["mqtth"]:
                kwargs["host"] = self.args["mqtth"]
            if self.args["namespace"]:
                kwargs["namespace"] = self.args["namespace"]
            if self.args["scene"]:
                kwargs["scene"] = self.args["scene"]

        if os.environ.get("SCENE"):
            self.scene = os.environ["SCENE"]
        elif "scene" in kwargs and kwargs["scene"]:
            if re.search("/", kwargs["scene"]):
                sys.exit("scene argument (scene) cannot include '/', aborting...")
            self.scene = kwargs["scene"]
            print("Cannot find SCENE environmental variable, using input parameter instead.")
        else:
            sys.exit("scene argument (scene) is unspecified or None, aborting...")
        super().__init__(
            realm,
            network_latency_interval,
            on_msg_callback,
            end_program_callback,
            video,
            debug,
            **kwargs
        )

        self.jitsi_host = self.config_data["ARENADefaults"]["jitsiHost"]
        self.persist_host = self.config_data["ARENADefaults"]["persistHost"]
        self.persist_path = self.config_data["ARENADefaults"]["persistPath"]

        self.persist_url = f"https://{self.persist_host}{self.persist_path}{self.namespaced_target}"

        # set up callbacks
        self.new_obj_callback = new_obj_callback
        self.delete_obj_callback = delete_obj_callback
        self.user_join_callback = user_join_callback
        self.user_left_callback = user_left_callback

        self.unspecified_object_ids = set() # objects that exist in the scene,
                                            # but this scene instance does not
                                            # have a reference to
        self.users = {} # dict of all users

        host = self.host.replace("mqtt.", "")
        print(f"Loading: https://{host}/{self.namespace}/{self.scene}, realm={self.realm}")

    def on_connect(self, client, userdata, flags, rc):
        super().on_connect(client, userdata, flags, rc)
        if rc == 0:
            # create ARENA-py Objects from persist server
            # no need to return anything here
            self.get_persisted_objs()

    async def process_message(self):
        while True:
            msg = await self.msg_queue.get()

            # extract payload
            try:
                payload_str = msg.payload.decode("utf-8", "ignore")
                payload = json.loads(payload_str)
            except Exception as e:
                print("Malformed payload, ignoring:")
                print(e)
                return

            try:
                # update object attributes, if possible
                if "object_id" in payload:
                    # parese payload
                    object_id = payload.get("object_id", None)
                    action = payload.get("action", None)

                    data = payload.get("data", {})
                    object_type = data.get("object_type", None)

                    event = None

                    # create/get object from object_id
                    if object_id in self.all_objects:
                        obj = self.all_objects[object_id]
                    else:
                        ObjClass = OBJECT_TYPE_MAP.get(object_type, Object)
                        obj = ObjClass(**payload)

                    # react to action accordingly
                    if action:
                        if action == "clientEvent":
                            event = Event(**payload)
                            if obj.evt_handler:
                                self.callback_wrapper(obj.evt_handler, event, payload)
                                continue

                        elif action == "delete":
                            if Camera.object_type in object_id: # object is a camera
                                if object_id in self.users:
                                    if self.user_left_callback:
                                        self.callback_wrapper(
                                                self.user_left_callback,
                                                self.users[object_id],
                                                payload
                                            )
                                    del self.users[object_id]
                            elif self.delete_obj_callback:
                                self.callback_wrapper(self.delete_obj_callback, obj, payload)
                            Object.remove(obj)
                            continue

                        else: # create/update
                            obj.update_attributes(**payload)

                    # call new message callback for all messages
                    if self.on_msg_callback:
                        if not event:
                            self.callback_wrapper(self.on_msg_callback, obj, payload)
                        else:
                            self.callback_wrapper(self.on_msg_callback, event, payload)

                    # run user_join_callback when user is found
                    if object_type and object_type == Camera.object_type:
                        if object_id not in self.users:
                            if object_id in self.all_objects:
                                self.users[object_id] = obj
                            else:
                                self.users[object_id] = Camera(**payload)

                            if self.user_join_callback:
                                self.callback_wrapper(
                                        self.user_join_callback,
                                        self.users[object_id],
                                        payload
                                    )

                    # if its an object the library has not seen before, call new object callback
                    elif object_id not in self.unspecified_object_ids and self.new_obj_callback:
                        self.callback_wrapper(self.new_obj_callback, obj, payload)
                        self.unspecified_object_ids.add(object_id)

            except Exception as e:
                print("Malformed message, ignoring:")
                print(e)

    def callback_wrapper(self, func, arg, msg):
        """Checks for number of arguments for callback"""
        if len(signature(func).parameters) != 3:
            print("[DEPRECATED]", "Callbacks and handlers now take 3 arguments: (scene, obj/evt, msg)!")
            func(arg)
        else:
            func(self, arg, msg)

    def generate_custom_event(self, evt, action="clientEvent"):
        """Publishes an custom event. Could be user or library defined"""
        return self._publish(evt, action)

    def generate_click_event(self, obj, type="mousedown", **kwargs):
        """Publishes an click event"""
        _type = type
        evt = Event(object_id=obj.object_id,
                    type=_type,
                    position=obj.data.position,
                    source=self.mqttc_id,
                    **kwargs)
        return self.generate_custom_event(evt, action="clientEvent")

    def manipulate_camera(self, cam, **kwargs):
        """Publishes a camera manipulation event"""
        if kwargs.get("position"):
            if isinstance(kwargs["position"], (list, tuple)):
                kwargs["position"] = Position(*kwargs["position"])
            elif isinstance(kwargs["position"], dict):
                kwargs["position"] = Position(**kwargs["position"])

        if kwargs.get("rotation"):
            if isinstance(kwargs["rotation"], (list, tuple)):
                kwargs["rotation"] = Rotation(*kwargs["rotation"])
            elif isinstance(kwargs["rotation"], dict):
                kwargs["rotation"] = Rotation(**kwargs["rotation"])

        if isinstance(cam, Object):
            object_id = cam.object_id
        else:
            object_id = cam
        evt = Event(object_id=object_id,
                    type="camera-override",
                    object_type=Camera.object_type,
                    **kwargs)
        return self.generate_custom_event(evt, action="update")

    def look_at(self, cam, target):
        """Publishes a camera manipulation event"""
        if isinstance(target, tuple) or isinstance(target, list):
            target = Position(*target)
        elif isinstance(target, dict):
            target = Position(**target)
        elif isinstance(target, Object):
            target = target.object_id

        if isinstance(cam, Object):
            object_id = cam.object_id
        else:
            object_id = cam
        evt = Event(object_id=object_id,
                    type="camera-override",
                    object_type="look-at",
                    target=target)
        return self.generate_custom_event(evt, action="update")

    @property
    def all_objects(self):
        """Returns all the objects in a scene"""
        return Object.all_objects

    def add_object(self, obj):
        """Public function to create an object"""
        res = self._publish(obj, "create")
        self.run_animations(obj)
        return res

    def add_objects(self, objs):
        """Public function to create multiple objects in a list"""
        for obj in objs:
            self.add_object(obj)
        return len(objs)

    def update_object(self, obj, **kwargs):
        """Public function to update an object"""
        if kwargs:
            obj.update_attributes(**kwargs)
        res = self._publish(obj, "update")
        self.run_animations(obj)
        return res

    def update_objects(self, objs, **kwargs):
        """Public function to update multiple objects in a list"""
        for obj in objs:
            self.update_object(obj, **kwargs)
        return len(objs)

    def delete_object(self, obj):
        """Public function to delete an object"""
        payload = {
            "object_id": obj.object_id
        }
        Object.remove(obj)
        return self._publish(payload, "delete")

    def run_animations(self, obj):
        """Runs all dispatched animations"""
        if obj.animations:
            payload = {
                "object_id": obj.object_id,
                "type": obj.type,
                "data": {"object_type": obj.data.object_type}
            }
            for i,animation in enumerate(obj.animations):
                if isinstance(animation, AnimationMixer):
                    payload["data"][f"animation-mixer"] = vars(animation)
                else:
                    anim = vars(animation).copy()
                    if i == 0:
                        payload["data"][f"animation"] = anim
                    else:
                        payload["data"][f"animation__{i}"] = anim
                    Utils.dict_key_replace(anim, "start", "from")
                    Utils.dict_key_replace(anim, "end", "to")
            obj.clear_animations()
            return self._publish(payload, "dispatch_animation")

    def _publish(self, obj, action):
        """Publishes to mqtt broker with "action":action"""
        topic = f"{self.root_topic}/{self.mqttc_id}/{obj['object_id']}"
        d = datetime.utcnow().isoformat()[:-3]+"Z"

        if action == "delete":
            payload = obj
            payload["action"] = "delete"
            payload["timestamp"] = d
            payload = json.dumps(payload)
        elif action == "dispatch_animation":
            payload = obj
            payload["action"] = "update"
            payload["timestamp"] = d
            payload = json.dumps(payload)
        else:
            payload = obj.json(action=action, timestamp=d)

        self.mqttc.publish(topic, payload, qos=0)
        if self.debug: print("[publish]", topic, payload)
        return payload

    def get_persisted_obj(self, object_id):
        """Returns a dictionary for a persisted object."""
        obj = None
        if object_id in self.all_objects:
            obj = self.all_objects[object_id]
            obj.persist = True
        else:
            # pass token to persist
            data = self.auth.urlopen(url=f"{self.persist_url}/{object_id}", creds=True)
            output = json.loads(data)
            if len(output) > 0:
                output = output[0]

                object_id = output["object_id"]
                data = output["attributes"]
                object_type = data.get("object_type", None)

                ObjClass = OBJECT_TYPE_MAP.get(object_type, Object)
                obj = ObjClass(object_id=object_id, data=data)
                obj.persist = True

        return obj

    def get_persisted_objs(self):
        """Returns a dictionary of persisted objects. [TODO] check object_type"""
        objs = {}
        # pass token to persist
        data = self.auth.urlopen(url=self.persist_url, creds=True)
        output = json.loads(data)
        for obj in output:
            if obj["type"] == Object.object_type or obj["type"] == Object.type:
                object_id = obj["object_id"]
                data = obj["attributes"]

                if object_id in self.all_objects:
                    persisted_obj = self.all_objects[object_id]
                    persisted_obj.persist = True
                else:
                    object_type = data.get("object_type", None)
                    ObjClass = OBJECT_TYPE_MAP.get(object_type, Object)
                    persisted_obj = ObjClass(object_id=object_id, data=data)
                    persisted_obj.persist = True

                objs[object_id] = persisted_obj

        return objs

    def get_persisted_scene_option(self):
        """Returns a dictionary for scene-options. [TODO] wrap the output as a BaseObject"""
        scene_opts_url = f"{self.persist_url}?type=scene-options"
        # pass token to persist
        data = self.auth.urlopen(url=scene_opts_url, creds=True )
        output = json.loads(data)
        return output

    def get_writable_scenes(self):
        """ Request list of scene names for logged in user account that user has publish permission for.
        Returns: list of scenes.
        """
        return self.auth.get_writable_scenes(host=self.host)

    def get_user_list(self):
        """Returns a list of users"""
        return self.users.values()

class Arena(Scene):
    """
    Another name for Scene.
    """



from .arena_object import Object

class Image(Object):
    """
    Class for Image in the ARENA.
    """
    object_type = "image"

    def __init__(self, url="", **kwargs):
        super().__init__(object_type=Image.object_type, url=url, **kwargs)



import json
import os
import re
import sys

from .arena_mqtt import ArenaMQTT


class Device(ArenaMQTT):
    """
    Gives access to an ARENA device.
    Can create and execute various user-defined functions/tasks.

    :param str host: Hostname of the MQTT broker (required).
    :param str realm: Reserved topic fork for future use (optional).
    :param str namespace: Username of authenticated user or other namespace (automatic).
    :param str device: The name of the device, without namespace (required).
    """

    def __init__(
                self,
                realm = "realm",
                network_latency_interval = 10000,  # run network latency update every 10s
                on_msg_callback = None,
                end_program_callback = None,
                debug = False,
                cli_args = False,
                **kwargs
            ):
        if cli_args:
            self.args = self.parse_cli()
            if self.args["mqtth"]:
                kwargs["host"] = self.args["mqtth"]
            if self.args["namespace"]:
                kwargs["namespace"] = self.args["namespace"]
            if self.args["device"]:
                kwargs["device"] = self.args["device"]

        if os.environ.get("DEVICE"):
            self.device = os.environ["DEVICE"]
        elif "device" in kwargs and kwargs["device"]:
            if re.search("/", kwargs["device"]):
                sys.exit("device argument (device) cannot include '/', aborting...")
            self.device = kwargs["device"]
            print("Cannot find DEVICE environmental variable, using input parameter instead.")
        else:
            sys.exit("device argument (device) is unspecified or None, aborting...")
        super().__init__(
            realm,
            network_latency_interval,
            on_msg_callback,
            end_program_callback,
            debug,
            **kwargs
        )
        print(f"Device topic ready: {self.realm}/d/{self.namespace}/{self.device}, host={self.host}")

    async def process_message(self):
        while True:
            msg = await self.msg_queue.get()
            # extract payload
            try:
                payload_str = msg.payload.decode("utf-8", "ignore")
                payload = json.loads(payload_str)
            except Exception as e:
                print("Malformed payload, ignoring:")
                print(e)
                return

    def publish(self, topic, payload_obj):
        """Publishes to mqtt broker."""
        payload = json.dumps(payload_obj)
        self.mqttc.publish(topic, payload, qos=0)
        if self.debug: print("[publish]", topic, payload)
        return payload



#!/usr/bin/env python3
"""
ARENA-py CLI

Usage: python3 -m arena -s <scene> -a <pub/sub> ...
"""

import argparse
import json

from arena import *

# define defaults
DEFAULT_MQTT_HOST = "mqtt.arenaxr.org"
PUBLISH = "pub"
SUBSCRIBE = "sub"


def on_msg_callback(scene, obj, msg):
    print(f"<{scene.root_topic}> \"{msg}\"")


def on_custom_topic_callback(client, userdata, msg):
    try:
        payload_str = msg.payload.decode("utf-8", "ignore")
        print(f"<{msg.topic}> \"{payload_str}\"")
    except:
        pass


def send_msg(scene, topic, msg):
    if topic is None:
        # use object topic name if possible
        try:
            json_msg = json.loads(msg)
            obj_topic = f"{scene.root_topic}/{json_msg['object_id']}"
        except:
            obj_topic = scene.root_topic
        print(f"Publishing to topic: <{obj_topic}>... ", end="")
        scene.mqttc.publish(obj_topic, msg)
    else:
        print(f"Publishing to topic: <{topic}>... ", end="")
        scene.mqttc.publish(topic, msg)
    print("done!")

    scene.stop_tasks()


def main(mqtth, realm, scene, namespace, action, topic, message):
    scene = Scene(host=mqtth, realm=realm, scene=scene, namespace=namespace)

    if action == SUBSCRIBE:
        if topic is None:
            print(f"Subscribing to topic: <{scene.subscribe_topic}>... ", end="")
            scene.on_msg_callback = on_msg_callback
        else:
            print(f"Subscribing to topic: <{topic}>... ", end="")
            scene.message_callback_add(topic, on_custom_topic_callback)
        print("done!")

    elif action == PUBLISH:
        if message is None:
            print("Message not specified! Aborting...")
            return
        scene.run_once(send_msg, scene=scene, topic=topic, msg=message)

    scene.run_tasks()


def cli():
    parser = argparse.ArgumentParser(description=("ARENA-py MQTT CLI"))

    parser.add_argument("-mh", "--mqtth",
                        type=str,
                        help="MQTT host to connect to",
                        default=DEFAULT_MQTT_HOST)
    parser.add_argument("-r", "--realm",
                        type=str,
                        help="Realm to listen to",
                        default="realm")
    parser.add_argument("-s", "--scene",
                        type=str,
                        help="Scene to listen to")
    parser.add_argument("-n", "--namespace",
                        type=str,
                        help="Namespace of scene",
                        default=None)

    parser.add_argument("-a", "--action",
                        help=f"Action to do ({PUBLISH} or {SUBSCRIBE})",
                        choices=(PUBLISH, SUBSCRIBE),
                        required=True)

    parser.add_argument("-t", "--topic",
                        type=str,
                        help=f"Custom topic to publish/subscribe to")
    parser.add_argument("-m", "--message",
                        type=str,
                        help=f"Message to send (ignored when action={SUBSCRIBE})")

    args = parser.parse_args()

    main(**vars(args))


if __name__ == "__main__":
    cli()



from .worker import Worker
import time

class PersistentWorker(Worker):
    """
    Wrapper for an ARENA task that runs at a periodic interval.
    """
    def __init__(self, event_loop, func, event=None, interval=1000, *args, **kwargs):
        super().__init__(event_loop, func, event, *args, **kwargs)
        self.interval = interval / 1000 # ms -> s

    async def run(self):
        if self.event: await self.event.wait()
        while True:
            start = time.time()
            try:
                self.func(*self.args, **self.kwargs)
            except Exception as e:
                func_name = self.func.__name__
                Worker.print_traceback(func_name)
                return
            end = time.time()
            target_time = max(0, self.interval - (end - start))
            await self.sleep(target_time)



import setuptools

with open("README.md", "r") as readme:
    long_description = readme.read()

setuptools.setup(
    name="arena-py",
    version="0.2.1",
    author="Conix Research Center",
    author_email="info@conix.io",
    license="BSD 3-clause \"New\" or \"Revised License\"",
    description="Draw objects and run programs in the ARENA using Python!",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/arenaxr/ARENA-py",
    packages=setuptools.find_packages(),
    entry_points = {
        'console_scripts': [
            'arena=arena.__main__:cli',
            'arena-py-permissions=arena.scripts.arena_py_permissions:main',
            'arena-py-pub=arena.scripts.arena_py_pub:main',
            'arena-py-signout=arena.scripts.arena_py_signout:main',
            'arena-py-sub=arena.scripts.arena_py_sub:main',
            'arena-py-token=arena.scripts.arena_py_token:main'
        ],
    },
    install_requires=[
        "aiohttp>=3.7.4",
        "paho-mqtt~=1.6.1",
        "requests~=2.28.1",
        "webcolors~=1.3",
        "google_auth_oauthlib~=0.5.2",
        "google-auth~=1.35.0",
        "PyJWT~=2.4.0",
        "numpy>=1.22",
        "scipy>=1.5.4"
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)



"""
auth.py - Authentication methods for accessing the ARENA.
"""

import datetime
import json
import os
import pickle
import ssl
import sys
import time
import webbrowser
from pathlib import Path
from urllib import parse, request
from urllib.error import HTTPError, URLError
from urllib.parse import urlsplit

import jwt
import requests
from google.auth import jwt as gJWT
from google.auth.transport.requests import AuthorizedSession, Request
from google_auth_oauthlib.flow import InstalledAppFlow

_gauth_file = ".arena_google_auth"
_mqtt_token_file = ".arena_mqtt_auth"
_arena_user_dir = f"{str(Path.home())}/.arena"
_local_mqtt_path = f"{_mqtt_token_file}"


class ArenaAuth:

    _scopes = ["openid",
               "https://www.googleapis.com/auth/userinfo.profile",
               "https://www.googleapis.com/auth/userinfo.email"]

    def __init__(self):
        self._csrftoken = None
        self._mqtt_token = None
        self._id_token = None

    def authenticate_user(self, host):
        """
        Begins authentication flow, getting Google auth, opening web browser if
        needed, getting username and state from ARENA server.
        host: The hostname of the ARENA webserver.
        Returns: Username from arena-account, or None.
        """
        print("Signing in to the ARENA...")
        gauth_json = self._get_gauthid(host)
        creds = None
        scene_auth_dir = self._get_scene_auth_path(host)
        scene_gauth_path = f"{scene_auth_dir}/{_gauth_file}"

        if not os.path.exists(scene_auth_dir):
            os.makedirs(scene_auth_dir)

        # store the user's access and refresh tokens
        if os.path.exists(scene_gauth_path):
            with open(scene_gauth_path, "rb") as token:
                creds = pickle.load(token)
            session = AuthorizedSession(creds)
            id_claims = gJWT.decode(creds.id_token, verify=False)

            # for reuse, client_id must still match
            gauth = json.loads(gauth_json)
            if "installed" not in gauth or "client_id" not in gauth["installed"]:
                creds = None
            if id_claims["aud"] != gauth["installed"]["client_id"]:
                creds = None
            if creds:
                print("Using cached Google authentication.")
        # if no credentials available, let the user log in.
        if not creds or not creds.valid:
            if creds and creds.expired and creds.refresh_token:
                print("Requesting refreshed Google authentication.")
                creds.refresh(Request())
                session = AuthorizedSession(creds)
            else:
                print("Requesting new Google authentication.")
                if self._is_headless_client():
                    # console flow for remote client
                    flow = InstalledAppFlow.from_client_config(
                        json.loads(gauth_json), self._scopes)
                    creds = flow.run_console()
                else:
                    # automated browser flow for local client
                    flow = InstalledAppFlow.from_client_config(
                        json.loads(gauth_json), self._scopes)
                    creds = flow.run_local_server(port=0)

                    # # alternate, run console flow with browser popup
                    # flow = InstalledAppFlow.from_client_config(
                    #     json.loads(gauth_json), self._scopes, redirect_uri="urn:ietf:wg:oauth:2.0:oob")
                    # auth_url, _ = flow.authorization_url(prompt="consent")
                    # print("Please go to this URL: {}".format(auth_url))
                    # webbrowser.open(auth_url, new=1, autoraise=True)
                    # code = input("Enter the authorization code: ")
                    # flow.fetch_token(code=code)
                    # creds = flow.credentials

                session = flow.authorized_session()
            with open(scene_gauth_path, "wb") as token:
                # save the credentials for the next run
                pickle.dump(creds, token)
            os.chmod(scene_gauth_path, 0o600)  # set user-only perms.

        username = None
        self._id_token = creds.id_token
        user_info = self._get_user_state(host, self._id_token)
        _user_info = json.loads(user_info)
        if "authenticated" in _user_info and "username" in _user_info:
            username = _user_info["username"]
        profile_info = session.get(
            "https://www.googleapis.com/userinfo/v2/me").json()
        if profile_info:
            print(f"Authenticated Google account: {profile_info['email']}")
        return username

    def authenticate_scene(self, host, realm, scene, username, video=False):
        """ End authentication flow, requesting permissions may change by owner
        or admin, for now, get a fresh mqtt_token each time.
        host: The hostname of the ARENA webserver.
        realm: The topic realm name.
        scene: The namespace/scene name combination.
        username: The ARENA username for the user.
        video: If Jitsi video conference is requested.
        Returns: username and mqtt_token from arena-account.
        """
        scene_auth_dir = self._get_scene_auth_path(host)
        scene_mqtt_path = f"{scene_auth_dir}/{_mqtt_token_file}"

        print("Using remote-authenticated MQTT token.")
        mqtt_json = self._get_mqtt_token(
            host, realm, scene, username, self._id_token, video)
        # save mqtt_token
        with open(scene_mqtt_path, mode="w") as d:
            d.write(mqtt_json)
        os.chmod(scene_mqtt_path, 0o600)  # set user-only perms.

        self._mqtt_token = json.loads(mqtt_json)
        self._log_token()
        return self._mqtt_token

    def authenticate_device(self, host):
        """
        Check for device mqtt_token, ask for a missing one, and save to local memory.
        """
        device_auth_dir = self._get_device_auth_path(host)
        device_mqtt_path = f"{device_auth_dir}/{_mqtt_token_file}"
        # check token expiration
        _remove_credentials(device_auth_dir, expire=True)
        # load device token if valid
        if os.path.exists(device_mqtt_path):
            print("Using user long-term device MQTT token.")
            f = open(device_mqtt_path, "r")
            mqtt_json = f.read()
            f.close()
        else:
            if not os.path.exists(device_auth_dir):
                os.makedirs(device_auth_dir)
            print(
                f"Generate a token for this device at https://{host}/user/profile")
            mqtt_json = input(
                "Paste auth MQTT full JSON here for this device: ")
            # save mqtt_token
            with open(device_mqtt_path, mode="w") as d:
                d.write(mqtt_json)
            os.chmod(device_mqtt_path, 0o600)  # set user-only perms.

        self._mqtt_token = json.loads(mqtt_json)
        self._log_token()
        return self._mqtt_token

    def _get_scene_auth_path(self, host):
        return f"{_arena_user_dir}/python/{host}/s"

    def _get_device_auth_path(self, host):
        return f"{_arena_user_dir}/python/{host}/d"

    def get_writable_scenes(self, host):
        """ Request list of scene names for logged in user that user has publish permission for.
        host: The hostname of the ARENA webserver.
        Returns: list of scenes.
        """
        my_scenes = self._get_my_scenes(host, self._id_token)
        return json.loads(my_scenes)

    def _is_headless_client(self):
        """ Determine headless or headed console.
        """
        if "SSH_TTY" in os.environ or "SSH_CLIENT" in os.environ:
            print("SSH connection detected, using headless auth.")
            return True
        try:
            webbrowser.get()
            return False
        except (webbrowser.Error) as err:
            print("Console-only OS detected. {0}".format(err))
            return True

    def _log_token(self):
        """
        Update user with token in use.
        """
        username = None
        if "username" in self._mqtt_token:
            username = self._mqtt_token["username"]
        print(f"ARENA Token Username: {username}")

        now = time.time()
        tok = jwt.decode(self._mqtt_token["token"], options={
                         "verify_signature": False})
        exp = float(tok["exp"])
        delta = (exp - now)
        dur_str = str(datetime.timedelta(milliseconds=delta*1000))
        print(f"ARENA Token valid for: {dur_str}h")

    def store_environment_auth(self, username, token):
        """
        Keep a copy of the token in local memory for urlopen and other tasks.
        """
        if username and token:
            print("Using environment MQTT token.")
            self._mqtt_token = {"username": username, "token": token}
            self._log_token()

    def check_local_auth(self):
        """
        Check for local mqtt_token and save to local memory.
        TODO: remove local check after ARTS supports mqtt_token passing
        """
        # 4 Oct 2021 remove deprecated user home creds path
        _remove_credentials(str(Path.home()))
        # check token expiration
        _remove_credentials(str(Path.cwd()), expire=True)
        # load local token if valid
        if os.path.exists(_local_mqtt_path):
            print("Using local MQTT token.")
            f = open(_local_mqtt_path, "r")
            mqtt_json = f.read()
            f.close()
            self._mqtt_token = json.loads(mqtt_json)
            self._log_token()
            return self._mqtt_token
        return None

    def _get_csrftoken(self, host):
        # get the csrftoken for django
        csrf_url = f"https://{host}/user/login"
        client = requests.session()
        client.get(csrf_url, verify=self.verify(host))  # sets cookie
        if "csrftoken" in client.cookies:
            self._csrftoken = client.cookies["csrftoken"]
        elif "csrf" in client.cookies:
            self._csrftoken = client.cookies["csrf"]
        else:
            self._csrftoken = None
        return self._csrftoken

    def _get_gauthid(self, host):
        url = f"https://{host}/conf/gauth.json"
        return self.urlopen(url)

    def _get_my_scenes(self, host, id_token):
        url = f"https://{host}/user/my_scenes"
        if not self._csrftoken:
            self._csrftoken = self._get_csrftoken(host)
        params = {"id_token": id_token}
        query_string = parse.urlencode(params)
        data = query_string.encode("ascii")
        return self.urlopen(url, data=data, csrf=self._csrftoken)

    def _get_user_state(self, host, id_token):
        url = f"https://{host}/user/user_state"
        if not self._csrftoken:
            self._csrftoken = self._get_csrftoken(host)
        params = {"id_token": id_token}
        query_string = parse.urlencode(params)
        data = query_string.encode("ascii")
        return self.urlopen(url, data=data, csrf=self._csrftoken)

    def _get_mqtt_token(self, host, realm, scene, username, id_token, video):
        url = f"https://{host}/user/mqtt_auth"
        if not self._csrftoken:
            self._csrftoken = self._get_csrftoken(host)
        params = {
            "id_auth": "google-installed",
            "username": username,
            "id_token": id_token,
            "realm": realm,
            "scene": scene
        }
        if video:
            params["camid"] = True
        query_string = parse.urlencode(params)
        data = query_string.encode("ascii")
        return self.urlopen(url, data=data, csrf=self._csrftoken)

    def verify(self, host):
        return host != "localhost"

    def urlopen(self, url, data=None, creds=False, csrf=None):
        """ urlopen is for ARENA URL connections.
        url: the url to POST/GET.
        data: None for GET, add params for POST.
        creds: True to pass the MQTT token as a cookie.
        csrf: The csrftoken.
        """
        urlparts = urlsplit(url)
        try:
            req = request.Request(url)
            if creds:
                req.add_header(
                    "Cookie", f"mqtt_token={self._mqtt_token['token']}")
            if csrf:
                req.add_header("Cookie", f"csrftoken={csrf}")
                req.add_header("X-CSRFToken", csrf)
            if self.verify(urlparts.netloc):
                res = request.urlopen(req, data=data)
            else:
                context = ssl.create_default_context()
                context.check_hostname = False
                context.verify_mode = ssl.CERT_NONE
                res = request.urlopen(req, data=data, context=context)
            return res.read().decode("utf-8")
        except (requests.exceptions.ConnectionError, ConnectionError, URLError, HTTPError) as err:
            print("{0}: ".format(err)+url)
            if isinstance(err, HTTPError) and round(err.code, -2) == 400:
                # user not authorized on website yet, they don"t have an ARENA username
                base_url = "{0.scheme}://{0.netloc}".format(urlsplit(url))
                print("Login with this this account on the website first:")
                print(f"Trying to open login page: {base_url}/user")
                try:
                    webbrowser.open_new_tab(f"{base_url}/user")
                except (webbrowser.Error) as err:
                    print("Console-only login. {0}".format(err))
            sys.exit("Terminating...")


def signout():
    for root, dirs, files in os.walk(_arena_user_dir):
        if _mqtt_token_file in files:
            _remove_credentials(root)
    if os.path.exists(_local_mqtt_path):
        _remove_credentials(_local_mqtt_path)
    print("Signed out of the ARENA.")


def _print_mqtt_token(storage_str, mqtt_claims):
    print("\nARENA MQTT/Video Permissions")
    print("----------------------")
    print(f"Storage: {storage_str}")
    print(f"User: {mqtt_claims['sub']}")
    exp_str = time.strftime("%c", time.localtime(mqtt_claims["exp"]))
    print(f"Expires: {exp_str}")
    if "room" in mqtt_claims:
        print("Video Conference: enabled")
    else:
        print("Video Conference: disabled")
    print("Publish topics:")
    for pub in mqtt_claims["publ"]:
        print(f"- {pub}")
    print("Subscribe topics:")
    for sub in mqtt_claims["subs"]:
        print(f"- {sub}")


def permissions():
    mqtt_claims = None
    # env storage auth
    if os.environ.get("ARENA_USERNAME") and os.environ.get("ARENA_PASSWORD"):
        mqtt_token = os.environ["ARENA_PASSWORD"]
        mqtt_claims = jwt.decode(mqtt_token["token"], options={
            "verify_signature": False})
        _print_mqtt_token(
            "environment variable 'ARENA_PASSWORD'", mqtt_claims)
    # file storage auth
    token_paths = []
    for root, dirs, files in os.walk(_arena_user_dir):
        if _mqtt_token_file in files:
            token_paths.append(os.path.join(root, _mqtt_token_file))
    if os.path.exists(_local_mqtt_path):
        token_paths.append(_local_mqtt_path)
    for mqtt_path in token_paths:
        f = open(mqtt_path, "r")
        mqtt_json = f.read()
        f.close()
        mqtt_token = json.loads(mqtt_json)
        mqtt_claims = jwt.decode(mqtt_token["token"], options={
            "verify_signature": False})
        _print_mqtt_token(mqtt_path, mqtt_claims)
    # no permissions
    if not mqtt_claims:
        print("Not signed into the ARENA.")


def _remove_credentials(cred_dir, expire=False):
    """
    Helper to remove credentials in path with expiration option.
    """
    test_gauth_path = f"{cred_dir}/{_gauth_file}"
    test_mqtt_path = f"{cred_dir}/{_mqtt_token_file}"
    if os.path.exists(test_mqtt_path):
        f = open(test_mqtt_path, "r")
        mqtt_json = f.read()
        f.close()
        mqtt_token = json.loads(mqtt_json)
        mqtt_claims = jwt.decode(mqtt_token["token"], options={
            "verify_signature": False})
        exp = float(mqtt_claims["exp"])
        now = time.time()
        if expire and now < exp:
            return  # exit if expire request is still good
        # otherwise remove
        os.remove(test_mqtt_path)
    if os.path.exists(test_gauth_path):
        os.remove(test_gauth_path)


if __name__ == "__main__":
    globals()[sys.argv[1]]()



from .arena_object import Object

class Icosahedron(Object):
    """
    Class for Icosahedron in the ARENA.
    """
    object_type = "icosahedron"

    def __init__(self, **kwargs):
        super().__init__(object_type=Icosahedron.object_type, **kwargs)
